// Win95, Win98, WinNT, Win2000, and WinXP Console API Functions. (WinCon.d) 
// Conversion from wincon.h to wincon.d done by David L. 'SpottedTiger' Davis

module std.c.windows.wincon;
  
private import std.c.windows.windows;
  
extern( Windows )
{

// * Attributes flags: *
enum
{
    FOREGROUND_BLUE            = 0x0001, // text color contains blue.
    FOREGROUND_GREEN           = 0x0002, // text color contains green.
    FOREGROUND_RED             = 0x0004, // text color contains red.
    FOREGROUND_INTENSITY       = 0x0008, // text color is intensified.
    BACKGROUND_BLUE            = 0x0010, // background color contains blue.
    BACKGROUND_GREEN           = 0x0020, // background color contains green.
    BACKGROUND_RED             = 0x0040, // background color contains red.
    BACKGROUND_INTENSITY       = 0x0080, // background color is intensified.
    COMMON_LVB_LEADING_BYTE    = 0x0100, // Leading Byte of DBCS
    COMMON_LVB_TRAILING_BYTE   = 0x0200, // Trailing Byte of DBCS
    COMMON_LVB_GRID_HORIZONTAL = 0x0400, // DBCS: Grid attribute: top horizontal.
    COMMON_LVB_GRID_LVERTICAL  = 0x0800, // DBCS: Grid attribute: left vertical.
    COMMON_LVB_GRID_RVERTICAL  = 0x1000, // DBCS: Grid attribute: right vertical.
    COMMON_LVB_REVERSE_VIDEO   = 0x4000, // DBCS: Reverse fore/back ground attribute.
    COMMON_LVB_UNDERSCORE      = 0x8000, // DBCS: Underscore.

    COMMON_LVB_SBCSDBCS        = 0x0300  // SBCS or DBCS flag.
}

// Console Screen Background Colors
// BLACK = 0, BLUE = 16, GREEN = 32 .. WHITE = 112 (step + 16)
enum BG_Color
{
    BLACK    = 0,   
    DKBLUE   = BACKGROUND_BLUE, 
    DKGREEN  = BACKGROUND_GREEN,
    DKAQUA   = BACKGROUND_BLUE  | BACKGROUND_GREEN,
    DKRED    = BACKGROUND_RED,
    DKPURPLE = BACKGROUND_BLUE  | BACKGROUND_RED,
    DKYELLOW = BACKGROUND_RED   | BACKGROUND_GREEN,
    DKWHITE  = BACKGROUND_BLUE  | BACKGROUND_RED   | BACKGROUND_GREEN,
    //-----------------------------------------------------------------         
    GRAY     = 0                | BACKGROUND_INTENSITY, 
    BLUE     = BACKGROUND_BLUE  | BACKGROUND_INTENSITY,
    GREEN    = BACKGROUND_GREEN | BACKGROUND_INTENSITY,
    AQUA     = BACKGROUND_BLUE  | BACKGROUND_GREEN | BACKGROUND_INTENSITY,
    RED      = BACKGROUND_RED   | BACKGROUND_INTENSITY,
    PURPLE   = BACKGROUND_BLUE  | BACKGROUND_RED   | BACKGROUND_INTENSITY,
    YELLOW   = BACKGROUND_RED   | BACKGROUND_GREEN | BACKGROUND_INTENSITY,
    WHITE    = BACKGROUND_BLUE  | BACKGROUND_RED   | BACKGROUND_GREEN | BACKGROUND_INTENSITY
};

// Console Screen Foreground Colors
// BLACK = 0 .. WHITE = 15
enum FG_Color
{
    BLACK    = 0,   
    DKBLUE   = FOREGROUND_BLUE, 
    DKGREEN  = FOREGROUND_GREEN,
    DKAQUA   = FOREGROUND_BLUE  | FOREGROUND_GREEN,
    DKRED    = FOREGROUND_RED,
    DKPURPLE = FOREGROUND_BLUE  | FOREGROUND_RED,
    DKYELLOW = FOREGROUND_RED   | FOREGROUND_GREEN,
    DKWHITE  = FOREGROUND_BLUE  | FOREGROUND_RED   | FOREGROUND_GREEN,
    //-----------------------------------------------------------------         
    GRAY     = 0                | FOREGROUND_INTENSITY, 
    BLUE     = FOREGROUND_BLUE  | FOREGROUND_INTENSITY,
    GREEN    = FOREGROUND_GREEN | FOREGROUND_INTENSITY,
    AQUA     = FOREGROUND_BLUE  | FOREGROUND_GREEN | FOREGROUND_INTENSITY,
    RED      = FOREGROUND_RED   | FOREGROUND_INTENSITY,
    PURPLE   = FOREGROUND_BLUE  | FOREGROUND_RED   | FOREGROUND_INTENSITY,
    YELLOW   = FOREGROUND_RED   | FOREGROUND_GREEN | FOREGROUND_INTENSITY,
    WHITE    = FOREGROUND_BLUE  | FOREGROUND_RED   | FOREGROUND_GREEN | FOREGROUND_INTENSITY
};

struct COORD 
{
    SHORT X;
    SHORT Y;
};

alias COORD *PCOORD;

struct SMALL_RECT 
{
    SHORT Left;
    SHORT Top;
    SHORT Right;
    SHORT Bottom;
};

alias SMALL_RECT *PSMALL_RECT;

struct KEY_EVENT_RECORD 
{
    BOOL bKeyDown;
    WORD wRepeatCount;
    WORD wVirtualKeyCode;
    WORD wVirtualScanCode;
    
    union uChar 
    {
        WCHAR UnicodeChar;
        CHAR   AsciiChar;
    }
    
    DWORD dwControlKeyState;
}; 

alias KEY_EVENT_RECORD *PKEY_EVENT_RECORD;

// * ControlKeyState flags *
enum
{
    RIGHT_ALT_PRESSED  = 0x0001,     // the right alt key is pressed.
    LEFT_ALT_PRESSED   = 0x0002,     // the left alt key is pressed.
    RIGHT_CTRL_PRESSED = 0x0004,     // the right ctrl key is pressed.
    LEFT_CTRL_PRESSED  = 0x0008,     // the left ctrl key is pressed.
    SHIFT_PRESSED      = 0x0010,     // the shift key is pressed.
    NUMLOCK_ON         = 0x0020,     // the numlock light is on.
    SCROLLLOCK_ON      = 0x0040,     // the scrolllock light is on.
    CAPSLOCK_ON        = 0x0080,     // the capslock light is on.
    ENHANCED_KEY       = 0x0100,     // the key is enhanced.
    NLS_DBCSCHAR       = 0x00010000, // DBCS for JPN: SBCS/DBCS mode.
    NLS_ALPHANUMERIC   = 0x00000000, // DBCS for JPN: Alphanumeric mode.
    NLS_KATAKANA       = 0x00020000, // DBCS for JPN: Katakana mode.
    NLS_HIRAGANA       = 0x00040000, // DBCS for JPN: Hiragana mode.
    NLS_ROMAN          = 0x00400000, // DBCS for JPN: Roman/Noroman mode.
    NLS_IME_CONVERSION = 0x00800000, // DBCS for JPN: IME conversion.
    NLS_IME_DISABLE    = 0x20000000  // DBCS for JPN: IME enable/disable.
}

struct MOUSE_EVENT_RECORD 
{
    COORD dwMousePosition;
    DWORD dwButtonState;
    DWORD dwControlKeyState;
    DWORD dwEventFlags;
};

alias MOUSE_EVENT_RECORD *PMOUSE_EVENT_RECORD;

// * ButtonState flags *
enum
{
    FROM_LEFT_1ST_BUTTON_PRESSED = 0x0001,
    RIGHTMOST_BUTTON_PRESSED     = 0x0002,
    FROM_LEFT_2ND_BUTTON_PRESSED = 0x0004,
    FROM_LEFT_3RD_BUTTON_PRESSED = 0x0008,
    FROM_LEFT_4TH_BUTTON_PRESSED = 0x0010
}

// * EventFlags *
enum
{
    MOUSE_MOVED   = 0x0001,
    DOUBLE_CLICK  = 0x0002,
    MOUSE_WHEELED = 0x0004
}

struct WINDOW_BUFFER_SIZE_RECORD 
{
    COORD dwSize;
};

alias WINDOW_BUFFER_SIZE_RECORD *PWINDOW_BUFFER_SIZE_RECORD;

struct MENU_EVENT_RECORD 
{
    UINT dwCommandId;
}; 

alias MENU_EVENT_RECORD *PMENU_EVENT_RECORD;

struct FOCUS_EVENT_RECORD 
{
    BOOL bSetFocus;
}; 

alias FOCUS_EVENT_RECORD *PFOCUS_EVENT_RECORD;

struct INPUT_RECORD 
{
    WORD EventType;
    union Event
    {
        KEY_EVENT_RECORD KeyEvent;
        MOUSE_EVENT_RECORD MouseEvent;
        WINDOW_BUFFER_SIZE_RECORD WindowBufferSizeEvent;
        MENU_EVENT_RECORD MenuEvent;
        FOCUS_EVENT_RECORD FocusEvent;
    }
};

alias INPUT_RECORD *PINPUT_RECORD;

// * EventType flags: *
enum
{
    KEY_EVENT                = 0x0001, // Event contains key event record
    MOUSE_EVENT              = 0x0002, // Event contains mouse event record
    WINDOW_BUFFER_SIZE_EVENT = 0x0004, // Event contains window change event record
    MENU_EVENT               = 0x0008, // Event contains menu event record
    FOCUS_EVENT              = 0x0010  // event contains focus change
}

struct CHAR_INFO 
{
    union UChar
    {
        WCHAR UnicodeChar;
        CHAR  AsciiChar;
    }

    UChar Char;
    WORD  Attributes;
};

alias CHAR_INFO *PCHAR_INFO;

struct CONSOLE_SCREEN_BUFFER_INFO 
{
    COORD dwSize;
    COORD dwCursorPosition;
    WORD  wAttributes;
    SMALL_RECT srWindow;
    COORD dwMaximumWindowSize;
};

alias CONSOLE_SCREEN_BUFFER_INFO *PCONSOLE_SCREEN_BUFFER_INFO; 

struct CONSOLE_CURSOR_INFO 
{
    DWORD  dwSize;
    BOOL   bVisible;
};

alias CONSOLE_CURSOR_INFO *PCONSOLE_CURSOR_INFO;

// * typedef for ctrl-c handler routines *
typedef BOOL (*PHANDLER_ROUTINE)
(
    DWORD CtrlType
);

enum
{
    CTRL_C_EVENT        = 0,
    CTRL_BREAK_EVENT    = 1,
    CTRL_CLOSE_EVENT    = 2,
    // 3 is reserved!
    // 4 is reserved!
    CTRL_LOGOFF_EVENT   = 5,
    CTRL_SHUTDOWN_EVENT = 6
}

//  * Input Mode flags: *
enum
{
    ENABLE_PROCESSED_INPUT = 0x0001,
    ENABLE_LINE_INPUT      = 0x0002,
    ENABLE_ECHO_INPUT      = 0x0004,
    ENABLE_WINDOW_INPUT    = 0x0008,
    ENABLE_MOUSE_INPUT     = 0x0010
}

// * Output Mode flags: *
enum
{
    ENABLE_PROCESSED_OUTPUT   = 0x0001,
    ENABLE_WRAP_AT_EOL_OUTPUT = 0x0002
}

// * Selection flags ( Win2000 & WinXP ) *
enum
{
    CONSOLE_NO_SELECTION          = 0x0000,
    CONSOLE_SELECTION_IN_PROGRESS = 0x0001,  // selection has begun
    CONSOLE_SELECTION_NOT_EMPTY   = 0x0002,  // non-null select rectangle
    CONSOLE_MOUSE_SELECTION       = 0x0004,  // selecting with mouse
    CONSOLE_MOUSE_DOWN            = 0x0008   // mouse is down
}

struct CONSOLE_SELECTION_INFO 
{  
    DWORD dwFlags;  
    COORD dwSelectionAnchor;  
    SMALL_RECT srSelection;
};

typedef CONSOLE_SELECTION_INFO *PCONSOLE_SELECTION_INFO;

// * direct API definitions. *
/+
 ' function reads data from the specified console input buffer without 
 ' removing it from the buffer.
 '
 +/
BOOL PeekConsoleInputA
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ PINPUT_RECORD lpBuffer,
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsRead
);

BOOL PeekConsoleInputW
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ PINPUT_RECORD lpBuffer,
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsRead
);

version( UNICODE )
{
    alias PeekConsoleInputW PeekConsoleInput;
}
else
{
    alias PeekConsoleInputA PeekConsoleInput;
} 

/+
 ' Function reads data from a console input buffer and 
 ' removes it from the buffer
 '
 +/
BOOL ReadConsoleInputA
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ PINPUT_RECORD lpBuffer,
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsRead
);

BOOL ReadConsoleInputW
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ PINPUT_RECORD lpBuffer,
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsRead
);

version( UNICODE )
{
    alias ReadConsoleInputW ReadConsoleInput;
}
else
{
    alias ReadConsoleInputA ReadConsoleInput;
}

/+
 ' Function writes data directly to the console input buffer.
 '
 +/
BOOL WriteConsoleInputA
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ IN  +/ INPUT_RECORD *lpBuffer,  
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsWritten
);

BOOL WriteConsoleInputW
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ IN  +/ INPUT_RECORD *lpBuffer, 
    /+ IN  +/ DWORD nLength,
    /+ OUT +/ LPDWORD lpNumberOfEventsWritten
);

version( UNICODE )
{
    alias WriteConsoleInputW WriteConsoleInput;
}
else
{
    alias WriteConsoleInputA WriteConsoleInput;
}

/+
 ' Function reads character and color attribute data from 
 ' a rectangular block of character cells in a console screen 
 ' buffer, and the function writes the data to a rectangular 
 ' block at a specified location in the destination buffer.
 '
 +/
BOOL ReadConsoleOutputA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ PCHAR_INFO lpBuffer,
    /+ IN  +/ COORD dwBufferSize,
    /+ IN  +/ COORD dwBufferCoord,
    /+ IN  +/ PSMALL_RECT lpReadRegion
);

BOOL ReadConsoleOutputW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ PCHAR_INFO lpBuffer,
    /+ IN  +/ COORD dwBufferSize,
    /+ IN  +/ COORD dwBufferCoord,
    /+ IN  +/ PSMALL_RECT lpReadRegion
);

version( UNICODE )
{
    alias ReadConsoleOutputW ReadConsoleOutput;
}
else
{
    alias ReadConsoleOutputA ReadConsoleOutput;
}

/+
 ' Function writes character and color attribute data to a specified 
 ' rectangular block of character cells in a console screen buffer. 
 ' The data to be written is taken from a correspondingly sized 
 ' rectangular block at a specified location in the source buffer.
 '
 +/
BOOL WriteConsoleOutputA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ CHAR_INFO *lpBuffer, 
    /+ IN  +/ COORD dwBufferSize,
    /+ IN  +/ COORD dwBufferCoord,
    /+ IN  +/ PSMALL_RECT lpWriteRegion
);

BOOL WriteConsoleOutputW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ CHAR_INFO *lpBuffer, 
    /+ IN  +/ COORD dwBufferSize,
    /+ IN  +/ COORD dwBufferCoord,
    /+ IN  +/ PSMALL_RECT lpWriteRegion
);

version( UNICODE )
{
    alias WriteConsoleOutputW WriteConsoleOutput;
}
else
{
    alias WriteConsoleOutputA WriteConsoleOutput;
}

/+
 ' Function copies a number of characters from consecutive cells of a 
 ' console screen buffer, beginning at a specified location.
 '
 +/
BOOL ReadConsoleOutputCharacterA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ LPSTR lpCharacter,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwReadCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsRead
);

BOOL ReadConsoleOutputCharacterW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ LPWSTR lpCharacter,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwReadCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsRead
);

version( UNICODE )
{
    alias ReadConsoleOutputCharacterW ReadConsoleOutputCharacter;
}
else
{
    alias ReadConsoleOutputCharacterA ReadConsoleOutputCharacter;
}

/+
 ' Function copies a specified number of character attributes from 
 ' consecutive cells of a console screen buffer, beginning at a 
 ' specified location.
 '
 +/
BOOL ReadConsoleOutputAttribute
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ LPWORD lpAttribute,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwReadCoord,
    /+ OUT +/ LPDWORD lpNumberOfAttrsRead
);

/+
 ' Function copies a number of characters to consecutive cells of a 
 ' console screen buffer, beginning at a specified location.
 '
 +/
BOOL WriteConsoleOutputCharacterA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ LPCSTR lpCharacter,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten
);

BOOL WriteConsoleOutputCharacterW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ LPCWSTR lpCharacter,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten
);

version( UNICODE )
{
    alias WriteConsoleOutputCharacterW WriteConsoleOutputCharacter;
}
else
{
    alias WriteConsoleOutputCharacterA WriteConsoleOutputCharacter;
}

/+
 ' Function copies a number of character attributes to consecutive 
 ' cells of a console screen buffer, beginning at a specified location.
 '
 +/
BOOL WriteConsoleOutputAttribute
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ WORD *lpAttribute, // CONST WORD *lpAttribute,
    /+ IN  +/ DWORD nLength,
    /+ IN  +/ COORD dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfAttrsWritten
);

/+
 ' Function writes a character to the console screen buffer a 
 ' specified number of times, beginning at the specified coordinates.
 '
 +/
BOOL FillConsoleOutputCharacterA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ CHAR  cCharacter,
    /+ IN  +/ DWORD  nLength,
    /+ IN  +/ COORD  dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten
);

BOOL FillConsoleOutputCharacterW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ WCHAR  cCharacter,
    /+ IN  +/ DWORD  nLength,
    /+ IN  +/ COORD  dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten
);

version( UNICODE )
{
    alias FillConsoleOutputCharacterW FillConsoleOutputCharacter;
}
else
{
    alias FillConsoleOutputCharacterA FillConsoleOutputCharacter;
}

/+
 ' Function sets the character attributes for a specified number of 
 ' character cells, beginning at the specified coordinates in a screen 
 ' buffer.
 '
 +/
BOOL FillConsoleOutputAttribute
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ WORD   wAttribute,
    /+ IN  +/ DWORD  nLength,
    /+ IN  +/ COORD  dwWriteCoord,
    /+ OUT +/ LPDWORD lpNumberOfAttrsWritten
);

/+
 ' Function retrieves the current input mode of a console's input 
 ' buffer or the current output mode of a console screen buffer.
 '
 +/
BOOL GetConsoleMode
(
    /+ IN  +/ HANDLE hConsoleHandle,
    /+ OUT +/ LPDWORD lpMode
);

/+
 ' Function retrieves the number of unread input records in the 
 ' console's input buffer.
 '
 +/
BOOL GetNumberOfConsoleInputEvents
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ LPDWORD lpNumberOfEvents
);

/+
 ' Function retrieves information about the specified console 
 ' screen buffer.
 '
 +/
BOOL GetConsoleScreenBufferInfo
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ PCONSOLE_SCREEN_BUFFER_INFO lpConsoleScreenBufferInfo
);

/+
 ' Function retrieves the size of the largest possible console window, 
 ' based on the current font and the size of the display.
 '
 +/
COORD GetLargestConsoleWindowSize
(
    /+ IN  +/ HANDLE hConsoleOutput
);

/+
 ' Function retrieves information about the size and visibility of 
 ' the cursor for the specified console screen buffer.
 '
 +/
BOOL GetConsoleCursorInfo
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ OUT +/ PCONSOLE_CURSOR_INFO lpConsoleCursorInfo
);

/+
 ' function retrieves the number of buttons on the mouse used by the 
 ' current console.
 '
 +/
BOOL GetNumberOfConsoleMouseButtons
(
    /+ IN  +/ LPDWORD lpNumberOfMouseButtons
);

/+
 ' Function sets the input mode of a console's input buffer or the 
 ' output mode of a console screen buffer.
 '
 +/
BOOL SetConsoleMode
(
    /+ IN  +/ HANDLE hConsoleHandle,
    /+ IN  +/ DWORD dwMode
);

/+
 ' Function sets the specified screen buffer to be the currently 
 ' displayed console screen buffer.
 '
 +/
BOOL SetConsoleActiveScreenBuffer
(
    /+ IN  +/ HANDLE hConsoleOutput
);

/+
 ' Function flushes the console input buffer. All input records 
 ' currently in the input buffer are discarded.
 '
 +/
BOOL FlushConsoleInputBuffer
(
    /+ IN  +/ HANDLE hConsoleInput
);

/+
 ' Function changes the size of the specified console screen buffer.
 '
 +/
BOOL SetConsoleScreenBufferSize
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ COORD dwSize
);

/+
 ' Function sets the cursor position in the specified console 
 ' screen buffer.
 '
 +/
BOOL SetConsoleCursorPosition
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ COORD dwCursorPosition
);

/+
 ' Function sets the size and visibility of the cursor for the 
 ' specified console screen buffer.
 '
 +/
BOOL SetConsoleCursorInfo
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ CONSOLE_CURSOR_INFO *lpConsoleCursorInfo  
);

/+
 ' Function moves a block of data in a screen buffer. The effects of 
 ' the move can be limited by specifying a clipping rectangle, so the 
 ' contents of the console screen buffer outside the clipping rectangle 
 ' are unchanged.
 '
 +/
BOOL ScrollConsoleScreenBufferA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ SMALL_RECT *lpScrollRectangle, 
    /+ IN  +/ SMALL_RECT *lpClipRectangle,   
    /+ IN  +/ COORD dwDestinationOrigin,
    /+ IN  +/ CHAR_INFO *lpFill             
);

BOOL ScrollConsoleScreenBufferW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ SMALL_RECT *lpScrollRectangle, 
    /+ IN  +/ SMALL_RECT *lpClipRectangle,   
    /+ IN  +/ COORD dwDestinationOrigin,
    /+ IN  +/ CHAR_INFO *lpFill             
);

version( UNICODE )
{
    alias ScrollConsoleScreenBufferW ScrollConsoleScreenBuffer;
}
else
{
    alias ScrollConsoleScreenBufferA ScrollConsoleScreenBuffer;
}

/+
 ' Function sets the current size and position of a console screen 
 ' buffer's window.
 '
 +/
BOOL SetConsoleWindowInfo
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ BOOL bAbsolute,
    /+ IN  +/ SMALL_RECT *lpConsoleWindow 
);

/+
 ' Function sets the attributes of characters written to the console 
 ' screen buffer by the WriteFile or WriteConsole function, or echoed 
 ' by the ReadFile or ReadConsole function. This function affects text 
 ' written after the function call.
 '
 +/
BOOL SetConsoleTextAttribute
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ WORD wAttributes
);

/+
 ' Function adds or removes an application-defined HandlerRoutine 
 ' function from the list of handler functions for the calling process.
 '
 ' If no handler function is specified, the function sets an inheritable 
 ' attribute that determines whether the calling process ignores CTRL+C 
 ' signals.
 '
 +/
BOOL SetConsoleCtrlHandler
(
    /+ IN  +/ PHANDLER_ROUTINE HandlerRoutine,
    /+ IN  +/ BOOL Add
);

/+
 ' Function sends a specified signal to a console process group that 
 ' shares the console associated with the calling process.
 '
 +/
BOOL GenerateConsoleCtrlEvent
(
    /+ IN  +/ DWORD dwCtrlEvent,
    /+ IN  +/ DWORD dwProcessGroupId
);

/+
 ' Function allocates a new console for the calling process.
 '
 +/ 
BOOL AllocConsole();

/+
 ' Function detaches the calling process from its console.
 '
 +/
BOOL FreeConsole();

/+
 ' Function retrieves the title bar string for the current 
 ' console window.
 '
 +/
DWORD GetConsoleTitleA
(
    /+ OUT +/ LPSTR lpConsoleTitle,
    /+ IN  +/ DWORD nSize
);

DWORD GetConsoleTitleW
(
    /+ OUT +/ LPWSTR lpConsoleTitle,
    /+ IN  +/ DWORD nSize
);

version( UNICODE )
{
    alias GetConsoleTitleW GetConsoleTitle;
}
else
{
    alias GetConsoleTitleA GetConsoleTitle;
}

/+
 ' Function sets the title bar string for the current console window.
 '
 +/
BOOL SetConsoleTitleA
(
    /+ IN  +/ LPCSTR lpConsoleTitle
);

BOOL SetConsoleTitleW
(
    /+ IN  +/ LPCWSTR lpConsoleTitle
);

version( UNICODE )
{
    alias SetConsoleTitleW SetConsoleTitle;
}
else
{
    alias SetConsoleTitleA SetConsoleTitle;
}

/+
 ' Function reads character input from the console input buffer 
 ' and removes it from the buffer.
 '
 +/
BOOL ReadConsoleA
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ LPVOID lpBuffer,           // less than 64Kb
    /+ IN  +/ DWORD nNumberOfCharsToRead,
    /+ OUT +/ LPDWORD lpNumberOfCharsRead,
    /+ IN  +/ LPVOID lpReserved
);

BOOL ReadConsoleW
(
    /+ IN  +/ HANDLE hConsoleInput,
    /+ OUT +/ LPVOID lpBuffer,
    /+ IN  +/ DWORD nNumberOfCharsToRead,
    /+ OUT +/ LPDWORD lpNumberOfCharsRead,
    /+ IN  +/ LPVOID lpReserved
);

version( UNICODE )
{
    alias ReadConsoleW ReadConsole;
}
else
{
    alias ReadConsoleA ReadConsole;
}

/+
 ' Function writes a character string to a console screen buffer 
 ' beginning at the current cursor location.
 ' 
 +/
BOOL WriteConsoleA
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ VOID *lpBuffer,  
    /+ IN  +/ DWORD nNumberOfCharsToWrite,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten,
    /+ IN  +/ LPVOID lpReserved  // Reserved; must be NULL
);

BOOL WriteConsoleW
(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ VOID *lpBuffer,  
    /+ IN  +/ DWORD nNumberOfCharsToWrite,
    /+ OUT +/ LPDWORD lpNumberOfCharsWritten,
    /+ IN  +/ LPVOID lpReserved // Reserved; must be NULL
);

version( UNICODE )
{
    alias WriteConsoleW WriteConsole;
}
else
{
    alias WriteConsoleA WriteConsole;
}

const int CONSOLE_TEXTMODE_BUFFER = 1;

HANDLE CreateConsoleScreenBuffer
(
    /+ IN  +/ DWORD dwDesiredAccess,
    /+ IN  +/ DWORD dwShareMode,
    /+ IN  +/ SECURITY_ATTRIBUTES *lpSecurityAttributes,  
    /+ IN  +/ DWORD dwFlags,
    /+ IN  +/ LPVOID lpScreenBufferData
);

/+
 ' Function retrieves the input code page used by the console 
 ' associated with the calling process. A console uses its input 
 ' code page to translate keyboard input into the corresponding 
 ' character value.
 +/
UINT GetConsoleCP();

/+
 ' Function sets the input code page used by the console associated 
 ' with the calling process. A console uses its input code page to 
 ' translate keyboard input into the corresponding character value.
 '
 +/
BOOL SetConsoleCP
(
    UINT wCodePageID
);

/+
 ' function retrieves the output code page used by the console 
 ' associated with the calling process. A console uses its output 
 ' code page to translate the character values written by the 
 ' various output functions into the images displayed in the console 
 ' window.
 '
 +/
UINT GetConsoleOutputCP();

/+
 ' Function sets the output code page used by the console associated 
 ' with the calling process. A console uses its output code page to 
 ' translate the character values written by the various output 
 ' functions into the images displayed in the console window.
 '
 ' HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage
 '
 +/
BOOL SetConsoleOutputCP
(
    /+ IN  +/ UINT wCodePageID
);

/+
 ' Function retrieves a handle for the standard input (STD_INPUT_HANDLE), 
 ' standard output (STD_OUTPUT_HANDLE), or standard error device (STD_ERROR_HANDLE).
 '
 +/
HANDLE GetStdHandle(
    /+ IN  +/ DWORD nStdHandle
);

/+
 ' Function sets the handle for the standard input(STD_INPUT_HANDLE), 
 ' standard output(STD_OUTPUT_HANDLE), or standard error device(STD_ERROR_HANDLE).
 '
 +/
BOOL SetStdHandle(
  /+ IN  +/ DWORD nStdHandle,
  /+ IN  +/ HANDLE hHandle
);

enum
{
    CONSOLE_FULLSCREEN          = 1, // fullscreen console
    CONSOLE_FULLSCREEN_HARDWARE = 2  // console owns the hardware
}

/+
 ' Function retrieves the display mode of the current console.
 ' ( Requires: WinXP )
 '
 +/
BOOL GetConsoleDisplayMode
(
    /+ OUT +/ LPDWORD lpModeFlags
);

/+
 ' Function retrieves the size of the font used by the specified 
 ' console screen buffer. ( Requires: WinXP )
 '
 +/
COORD GetConsoleFontSize(
    /+ IN  +/ HANDLE hConsoleOutput,
    /+ IN  +/ DWORD nFont
);

/+
 ' Function retrieves the window handle used by the console 
 ' associated with the calling process.
 '
 +/
HWND GetConsoleWindow();

/+
 ' Function retrieves a list of the processes attached to the 
 ' current console.
 '
 +/
DWORD GetConsoleProcessList
(
    /+ OUT +/ LPDWORD lpdwProcessList, // Less-Than 64Kb
    /+ IN  +/ DWORD dwProcessCount
);

} // end extern( Windows )
